"""Created on Oct 21 00:12:02 2024"""


class PlotError(Exception):
    """Basic PlotError class"""
    pass


class NoXYLabels(PlotError):
    """Custom class for missing x or y labels"""
    pass


class OrientationError(PlotError):
    """Custom class for wrong orientation"""
    pass



"""Created on Jul 23 16:56:48 2022."""

from .backend.uIO import check_for_errors


def read_txt_file(file_to_read: str) -> list:
    """
    Reads a text file and returns its contents as a list of stripped lines.

    This function reads the specified text file line by line, strips any newline characters,
    and returns a list of those lines.

    Parameters
    ----------
    file_to_read : str
        The path to the text file to be read.

    Returns
    -------
    list of str
        A list containing each line from the file, with newline characters removed.

    Raises
    ------
    FileNotFoundError
        If the file specified by `file_to_read` does not exist or cannot be accessed.
    IOError
        If there is an issue reading the file.
    """
    with open(file_to_read, 'r') as file:
        return [line.strip() for line in file]


def get_lines_from_txt_file(file_to_read: str, lines_to_read: list) -> list:
    """
    Retrieves specific lines from a text file.

    This function reads a file and returns only the specified lines from the file.
    If `lines_to_read` is a single integer, it is converted to a list. The function
    uses zero-based indexing for line numbers.

    Parameters
    ----------
    file_to_read : str
        The path to the text file to be read.
    lines_to_read : list of int or int
        The line numbers (zero-based) to retrieve from the file. If a single integer is provided,
        it will be treated as a list with one element.

    Returns
    -------
    list of str
        A list of the lines from the file that correspond to the provided line numbers.

    Raises
    ------
    ValueError
        If any of the requested line numbers exceed the total number of lines in the file.
    TypeError
        If `lines_to_read` is not an integer or a list of integers.
    """
    out = read_txt_file(file_to_read)

    if not isinstance(lines_to_read, list):
        lines_to_read = [lines_to_read]

    if not all(isinstance(line, int) for line in lines_to_read):
        raise TypeError("lines_to_read must be an integer or a list of integers.")

    check_for_errors(open_file=out, lines_to_read=lines_to_read)

    # taken from https://stackoverflow.com/a/2082131/3212945
    return [value for index, value in enumerate(out) if index in set(lines_to_read)]



"""Created on Jul 20 21:45:16 2022."""

from copy import deepcopy
from itertools import compress
from typing import List, Union

from . import eList as eL
from .eList import GotAnUnknownValue, IndexOutOfList, UnequalElements


def equalizing_list_length(primary_list, secondary_list, names):
    primary_len = len(primary_list)
    secondary_len = len(secondary_list)

    if secondary_len > primary_len:
        raise UnequalElements(f'The number of elements in the {names[0]} list is greater than that of {names[1]}. '
                              f'Cannot perform replacement in this case.')
    elif secondary_len < primary_len:
        secondary_list += secondary_list[:primary_len - secondary_len]

    return secondary_list


def replace_at_index(input_list, index, value, new_list=False):
    index, value = map(lambda x: [x] if not isinstance(x, list) else x, (index, value))
    value += value[:len(index) - len(value)]

    for i, v in zip(index, value):
        if i > len(input_list) - 1:
            raise IndexOutOfList(f'Index {i} is out of bound for a list of length {len(input_list)}.')

    if new_list:
        input_list = input_list[:]

    for i, v in zip(index, value):
        input_list[i] = v

    return input_list


def replace_element(input_list, old_element, new_element, new_list=False):
    old_element, new_element = map(lambda x: [x] if not isinstance(x, list) else x, (old_element, new_element))
    new_element += new_element[:len(old_element) - len(new_element)]

    index = []
    for i, old in enumerate(old_element):
        if old not in input_list:
            raise GotAnUnknownValue(f'The value {old} given in old_element does not exist in the input_list.')
        index.append(input_list.index(old))

    if new_list:
        input_list = input_list[:]

    for i, new in zip(index, new_element):
        input_list[i] = new

    return input_list


class CountObjectsInList:

    def __init__(self, counter_dict):
        self.counter_dict = counter_dict
        self.__counter_dict = sorted(self.counter_dict.items(), key=lambda x: x[1], reverse=True)

        self.counter = 0

    def __str__(self):
        out = '-' * 50 + '\n'
        out += f'|{"items":^30}|{"counts":^17}|\n'
        out += '-' * 50 + '\n'
        out += '\n'.join([f'|{key:^30}|{value:^17}|'
                          if not isinstance(key, str)
                          else f"|\'{key}\':^30|{value:^17}|"
                          for key, value in self.counter_dict.items()]) + '\n'
        out += '-' * 50 + '\n'
        return out

    def __getitem__(self, item):
        _get = self.__counter_dict[item]
        try:
            return CountObjectsInList({element[0]: element[1] for element in _get})
        except TypeError:
            return CountObjectsInList({_get[0]: _get[1]})


def numeric_list_to_string(num_list: List[int]) -> List[str]:
    """
    Convert all elements of a numeric lists to string.

    Parameters
    ----------
    num_list : List[int]
        A list containing numeric values.

    Returns
    -------
    List[str]
        A list containing strings equivalent to the numbers in the input list.

    Examples
    --------
    Let's say we have

    >>> a = [1, 2, 3, 4, 5]

    and we want to convert each element in thist list to string,

    >>> numeric_list_to_string(a)
    >>> ['1', '2', '3', '4', '5']
    """
    return list(map(str, num_list))


class Replace:
    """Class to replace stuff inside a given list."""

    def __init__(self, input_list: list, work_on: Union[list, int], replace_with: Union[list, int],
                 new_list: bool = False, by: str = 'index'):
        self.input_list = deepcopy(input_list) if new_list else input_list
        self.work_on = work_on
        self.replace_with = replace_with
        self.by = by

    def __convert_inputs_to_lists(self):
        if not isinstance(self.work_on, list):
            self.work_on = [self.work_on]

        if not isinstance(self.replace_with, list):
            self.replace_with = [self.replace_with]

        return self.work_on, self.replace_with

    def __equalizing_list_length(self) -> list:
        if self.by == 'index':
            names = ['index', 'value']
        elif self.by == 'value':
            names = ['old_elements', 'new_elements']
        else:
            raise eL.InvalidInputParameter('The input parameter required is, \'index\', or \'value\'.')

        if len(self.replace_with) != len(self.work_on):
            raise eL.UnequalElements(f'The number of elements in {names[0]} list is not equal to that of {names[1]}.'
                                     f' Cannot perform replacement in this case.')

        return self.replace_with

    def __replace_values(self, primary_list=None) -> None:
        primary_list = self.work_on if primary_list is None else primary_list

        for index, value in zip(primary_list, self.replace_with):
            self.input_list[index] = value

    def at_index(self) -> list:
        """
        Replaces the elements on the specified indices.

        Returns
        -------
        list:
            A list with replaced values.
        """
        self.work_on, self.replace_with = self.__convert_inputs_to_lists()
        self.replace_with = self.__equalizing_list_length()

        bool_mask = [_index > len(self.input_list) - 1 for _index in self.work_on]

        if any(bool_mask):
            join_ = ", ".join(compress(numeric_list_to_string(self.replace_with), bool_mask))
            raise eL.IndexOutOfList(f'Index {join_} is out of bound for a list of length {len(self.input_list)}.')

        self.__replace_values()

        return self.input_list

    def at_value(self) -> list:
        """
        Replaces the specified values in the given input list.

        Returns
        -------
        list:
            A list with replaced values.
        """
        self.work_on, self.replace_with = self.__convert_inputs_to_lists()
        self.replace_with = self.__equalizing_list_length()

        bool_mask = [element not in self.input_list for element in self.work_on]

        if any(bool_mask):
            join_ = ", ".join(compress(numeric_list_to_string(self.work_on), bool_mask))
            raise eL.GotAnUnknownValue(f'The value {join_} given in old_element does not exist in the input_list.')

        index = [self.input_list.index(element) for element in self.work_on]

        self.__replace_values(primary_list=index)

        return self.input_list



"""Created on Oct 29 09:33:06 2024"""

__all__ = ['LinePlot', 'ScatterPlot', 'SubPlots', 'label_handler', 'plot_or_scatter', 'plot_dictionary_handler',
           'split_dictionary', 'dual_axes_data_validation', 'dual_axes_label_management']

import warnings
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from matplotlib import pyplot as plt, rcParams

from .ePlotting import NoXYLabels

# SAFEGUARDS:
rc_color = rcParams['axes.prop_cycle'].by_key()['color'] * 10
_split = Tuple[Union['LinePlot', 'ScatterPlot'], Union['LinePlot', 'ScatterPlot']]
label_management = Tuple[str, str, str, str, List[str]]


class _PlotParams:
    """
    Base class for handling common plot parameters.

    Parameters
    ----------
    line_style : list(str), optional
        Line style for the plot. Default is None.
    line_width : list(float), optional
        Width of the lines. Default is None.
    color : list(str), optional
        List or tuple containing two colors. Defaults to first two colors in matplotlib color cycle.
    alpha : list(float), optional
        Opacity of the plot elements. Default is None.
    marker : list(str), optional
        Marker style for the plot. Default is None.
    marker_size : list(float), optional
        Size of the markers. Default is None.
    marker_edge_color : list(str), optional
        Color of the marker edges. Default is None.
    marker_face_color : list(str), optional
        Color of the marker faces. Default is None.
    marker_edge_width : list(float), optional
        Width of the marker edges. Default is None.
    size : list(float), optional
        Size of the markers for scatter plots. Must be a list or tuple with exactly two elements. Default is None.
    cmap : list(Colormap), optional
        Colormap used for scatter plots. Default is None.
    face_color : list(str), optional
        Color of the marker faces in scatter plots. Default is None.
    """
    
    def __init__(self,
                 line_style=None, line_width=None,
                 color=None, alpha=None,
                 marker=None, marker_size=None,
                 marker_edge_color=None, marker_face_color=None, marker_edge_width=None,
                 size=None, cmap=None, face_color=None,
                 share_x=None, share_y=None, subplot_fig_size=None):
        
        self.line_style = line_style
        self.line_width = line_width
        self.color = color
        self.alpha = alpha
        self.marker = marker
        self.marker_size = marker_size
        self.marker_edge_color = marker_edge_color
        self.marker_face_color = marker_face_color
        self.marker_edge_width = marker_edge_width
        
        # Additional keywords for scatter plot
        self.size = size
        self.cmap = cmap
        self.face_color = face_color
        
        # Additional keywords for subplots
        self.share_x = share_x
        self.share_y = share_y
        self.fig_size = subplot_fig_size
    
    def to_dict(self) -> dict:
        """
        Convert the plot parameters to a dictionary.

        Returns
        -------
        dict
            A dictionary where keys are parameter labels and values are the corresponding
            plot parameters. Parameters that are `None` are also included in the dictionary.
        """
        return {label: param for label, param in zip(self._all_labels(), self._all_parameters())}
    
    def get(self):
        """
        Get the dictionary of plot parameters, excluding None values.

        Returns
        -------
        dict
            Dictionary containing non-None parameters for the plot.
        """
        param_dict = {}
        
        for param, label in zip(self._all_parameters(), self._all_labels()):
            if param is not None:
                param_dict[f'{label}'] = param
        
        return param_dict
    
    def _all_parameters(self):
        raise NotImplementedError("This method should be implemented by subclasses.")
    
    def _all_labels(self):
        raise NotImplementedError("This method should be implemented by subclasses.")


class LinePlot(_PlotParams):
    """
    Class for double-line plot parameters.

    Parameters
    ----------
    All parameters are inherited from `_PlotParams`.
    """
    
    def __init__(self, line_style=None, line_width=None, color=None, alpha=None,
                 marker=None, marker_size=None, marker_edge_color=None,
                 marker_face_color=None, marker_edge_width=None, _fixed: int = 0):
        super().__init__(line_style=line_style, line_width=line_width, color=color,
                         alpha=alpha, marker=marker, marker_size=marker_size,
                         marker_edge_color=marker_edge_color, marker_face_color=marker_face_color,
                         marker_edge_width=marker_edge_width)
        self.color = color or rc_color
    
    def __repr__(self):
        param_str = ', '.join(f"{key}={value!r}" for key, value in self.to_dict().items())
        return f"{self.__class__.__name__}({param_str})"
    
    def __eq__(self, other):
        if not isinstance(other, LinePlot):
            return NotImplemented
        return self.to_dict() == other.to_dict()
    
    def __hash__(self):
        # Hash based on the tuple of sorted key-value pairs in the dictionary
        return hash(tuple(sorted(self.to_dict().items())))
    
    @classmethod
    def populate(cls, dictionary: Dict[str, Any]) -> 'LinePlot':
        """
        Create an instance of `LinePlot` from a dictionary of parameters.

        Parameters
        ----------
        dictionary : dict
            A dictionary where keys represent parameter labels (e.g., 'ls' for line style, 'lw' for line width),
            and values represent the corresponding values for each parameter.

        Returns
        -------
        LinePlot
            An instance of `LinePlot` with attributes populated based on the provided dictionary.
        """
        instance = cls()
        for key, value in dictionary.items():
            attr_name = {'ls': 'line_style',
                         'lw': 'line_width',
                         'color': 'color',
                         'alpha': 'alpha',
                         'marker': 'marker',
                         'ms': 'marker_size',
                         'mec': 'marker_edge_color',
                         'mfc': 'marker_face_color',
                         'mew': 'marker_edge_width'}.get(key, key)
            setattr(instance, attr_name, value)
        
        return instance
    
    def _all_parameters(self):
        return [self.line_style, self.line_width,
                self.color, self.alpha,
                self.marker, self.marker_size,
                self.marker_edge_color, self.marker_face_color, self.marker_edge_width]
    
    def _all_labels(self):
        return ['ls', 'lw', 'color', 'alpha', 'marker', 'ms', 'mec', 'mfc', 'mew']


class ScatterPlot(_PlotParams):
    """
    Class for double-scatter plot parameters.

    Parameters
    ----------
    All parameters are inherited from `_PlotParams`.
    """
    
    def __init__(self, color=None, alpha=None, marker=None, size=None, cmap=None, face_color=None):
        super().__init__(color=color, alpha=alpha, marker=marker, size=size, cmap=cmap, face_color=face_color)
        
        if self.color is None:
            self.color = rc_color
    
    def __repr__(self):
        param_str = ', '.join(f"{key}={value!r}" for key, value in self.to_dict().items())
        return f"{self.__class__.__name__}({param_str})"
    
    def __eq__(self, other):
        if not isinstance(other, LinePlot):
            return NotImplemented
        return self.to_dict() == other.to_dict()
    
    def __hash__(self):
        # Hash based on the tuple of sorted key-value pairs in the dictionary
        return hash(tuple(sorted(self.to_dict().items())))
    
    @classmethod
    def populate(cls, dictionary: Dict[str, Any]) -> 'ScatterPlot':
        """
        Create an instance of `ScatterPlot` from a dictionary of parameters.

        Parameters
        ----------
        dictionary : dict
            A dictionary where keys represent parameter labels (e.g., 's' for size, 'c' for color), and values represent the corresponding values for each parameter.

        Returns
        -------
        ScatterPlot
            An instance of `ScatterPlot` with attributes populated based on the provided dictionary.
        """
        instance = cls()
        for key, value in dictionary.items():
            attr_name = {'c': 'color',
                         'alpha': 'alpha',
                         'marker': 'marker',
                         's': 'size',
                         'cmap': 'cmap',
                         'fc': 'face_color'}.get(key, key)
            setattr(instance, attr_name, value)
        
        return instance
    
    def _all_parameters(self):
        return [self.color, self.alpha, self.marker, self.size, self.cmap, self.face_color]
    
    def _all_labels(self):
        return ['c', 'alpha', 'marker', 's', 'cmap', 'fc']


class SubPlots(_PlotParams):
    
    def __init__(self, share_x=None, share_y=None, fig_size=None):
        super().__init__(share_x=share_x, share_y=share_y, subplot_fig_size=fig_size)
    
    def _all_labels(self):
        return ['sharex', 'sharey', 'figsize']
    
    def _all_parameters(self):
        return [self.share_x, self.share_y, self.fig_size]


def label_handler(x_labels: Optional[List[str]], y_labels: Optional[List[str]],
                  n_rows: int, n_cols: int, auto_label: bool) -> Tuple[List[str], List[str]]:
    """
    Handles the generation or validation of x and y labels for a subplot configuration.

    Parameters
    ----------
    x_labels : list of str or None
        The labels for the x-axis. If `None`, labels may be auto-generated based on `auto_label`.
    y_labels : list of str or None
        The labels for the y-axis. If `None`, labels may be auto-generated based on `auto_label`.
    n_rows : int
        Number of rows in the subplot grid.
    n_cols : int
        Number of columns in the subplot grid.
    auto_label : bool
        If `True`, generates x and y labels automatically when not provided.

    Returns
    -------
    Tuple[List[str], List[str]]
        The x and y labels for the subplot grid.

    Raises
    ------
    NoXYLabels
        If both `x_labels` and `y_labels` are `None` and `auto_label` is `False`.

    Warnings
    --------
    UserWarning
        If one of `x_labels` or `y_labels` is missing when `auto_label` is enabled, or if
        there is a mismatch in the number of provided labels.

    """
    if not auto_label and (x_labels is None or y_labels is None):
        raise NoXYLabels("Both x_labels and y_labels are required without the auto_label parameter.")
    
    elif auto_label and (x_labels is None or y_labels is None):
        if x_labels is None and y_labels is None:
            pass
        else:
            if x_labels is None:
                warnings.warn("y_labels given but x_labels is missing, applying auto-labeling...", UserWarning)
            if y_labels is None:
                warnings.warn("x_labels given but y_labels is missing, applying auto-labeling...", UserWarning)
    
    if auto_label:
        if x_labels and y_labels:
            start = "auto_label selected with x_labels and y_labels provided"
            if len(x_labels) != n_rows * n_cols or len(y_labels) != n_rows * n_cols:
                warnings.warn(f"{start}, mismatch found, using auto-generated labels...", UserWarning)
                x_labels = [fr'X$_{i + 1}$' for i in range(n_cols * n_rows)]
                y_labels = [fr'Y$_{i + 1}$' for i in range(n_cols * n_rows)]
            else:
                print(f"{start}, using user-provided labels...")
        else:
            x_labels = [fr'X$_{i + 1}$' for i in range(n_cols * n_rows)]
            y_labels = [fr'Y$_{i + 1}$' for i in range(n_cols * n_rows)]
    
    return x_labels, y_labels


def plot_or_scatter(axes: plt.axis, scatter: bool):
    """
    Returns the plot or scatter method based on the specified plot type.

    Parameters
    ----------
    axes : plt.axis
        The matplotlib axis on which to apply the plot or scatter method.
    scatter : bool
        If True, returns the scatter method; otherwise, returns the plot method.

    Returns
    -------
    function
        The matplotlib plotting method (`axes.scatter` if scatter is True, otherwise `axes.plot`).
    """
    return axes.scatter if scatter else axes.plot


def plot_dictionary_handler(plot_dictionary: Union[LinePlot, ScatterPlot]):
    """
    Handles plot dictionary configuration, retrieving items from the specified plot dictionary.

    If no dictionary is provided, returns items from a default LinePlot instance.

    Parameters
    ----------
    plot_dictionary : Union[LinePlot, ScatterPlot]
        The plot dictionary to retrieve items from. If None, defaults to a LinePlot instance.

    Returns
    -------
    Iterable
        An iterable of items (key-value pairs) from the specified or default plot dictionary.
    """
    return plot_dictionary.get().items() if plot_dictionary else LinePlot().get().items()


def split_dictionary(plot_instance: Union[LinePlot, ScatterPlot]) -> _split:
    """
    Split a `LinePlot` or `ScatterPlot` instance's parameters into two separate instances of the same type.

    Parameters
    ----------
    plot_instance : Union[LinePlot, ScatterPlot]
        An instance of `LinePlot` or `ScatterPlot` with parameters stored as lists or tuples of two elements.
        Each parameter should be a list or tuple containing exactly two values, corresponding to settings
        for the two resulting instances.

    Returns
    -------
    Tuple[Union[LinePlot, ScatterPlot], Union[LinePlot, ScatterPlot]]
        Two instances of the same type as `plot_instance` (either `LinePlot` or `ScatterPlot`),
        with parameters split based on the values in `plot_instance`. The first instance (`instance1`)
        and second instance (`instance2`) will have their attributes set according to the first and second
        elements, respectively, from each list or tuple in `plot_instance`.

    Raises
    ------
    ValueError
        If any parameter in `plot_instance` is not a list or tuple with exactly two elements.
    """
    # Flatten the parameters from the input plot instance
    parameters = plot_instance.get()
    params_instance1, params_instance2 = {}, {}
    
    # Split each parameter into two separate dictionaries for the two instances
    for param_name, values in parameters.items():
        if isinstance(values, (list, tuple)) and len(values) == 2:
            params_instance1[param_name], params_instance2[param_name] = values
        else:
            raise ValueError(f"Parameter '{param_name}' must be a list or tuple with exactly two elements.")
    
    instance1 = plot_instance.__class__.populate(params_instance1)
    instance2 = plot_instance.__class__.populate(params_instance2)
    
    return instance1, instance2


def dual_axes_data_validation(x1_data: np.ndarray, x2_data: Optional[np.ndarray],
                              y1_data: np.ndarray, y2_data: Optional[np.ndarray],
                              use_twin_x: bool, axis_labels: Union[List[str], str]) -> None:
    """
    Validates the data and parameters for dual-axes plotting.

    Parameters
    ----------
    x1_data : np.ndarray
        Data for the primary x-axis.
    x2_data : Optional[np.ndarray]
        Data for the secondary x-axis (used in dual x-axis plots). Should be `None` if `use_twin_x` is True.
    y1_data : np.ndarray
        Data for the primary y-axis.
    y2_data : Optional[np.ndarray]
        Data for the secondary y-axis (used in dual y-axis plots). Should be `None` if `use_twin_x` is False.
    use_twin_x : bool
        If True, a dual y-axis plot is expected; otherwise, a dual x-axis plot is expected.
    axis_labels : List[str]
        List of axis labels. Must have exactly three elements:
        - Label for the x-axis of the primary plot.
        - Label for the y-axis of the primary plot.
        - Label for the secondary axis (x or y).

    Raises
    ------
    ValueError
        If any of the following conditions are met:
        - `axis_labels` does not have exactly three elements.
        - `x1_data` or `y1_data` is empty.
        - `x2_data` is provided when `use_twin_x` is True.
        - `y2_data` is provided when `use_twin_x` is False.
    """
    if len(axis_labels) != 3:
        raise ValueError("The axis_labels should have a length of 3.")
    if len(x1_data) == 0 or len(y1_data) == 0:
        raise ValueError("Primary x or y data is empty. Please provide valid data.")
    if use_twin_x and x2_data is not None:
        raise ValueError("Dual Y-axis plot requested but 'x2_data' given.")
    if not use_twin_x and y2_data is not None:
        raise ValueError("Dual X-axis plot requested but 'y2_data' given.")


def dual_axes_label_management(x1y1_label: Optional[str], x1y2_label: Optional[str], x2y1_label: Optional[str],
                               auto_label: bool, axis_labels: Optional[List[str]], plot_title: Optional[str],
                               use_twin_x: bool) -> label_management:
    """
    Manages labels and titles for dual-axes plots, with options for automatic labeling.

    Parameters
    ----------
    x1y1_label : Optional[str]
        Label for the primary plot (X1 vs Y1).
    x1y2_label : Optional[str]
        Label for the secondary Y-axis plot (X1 vs Y2), used if `use_twin_x` is True.
    x2y1_label : Optional[str]
        Label for the secondary X-axis plot (X2 vs Y1), used if `use_twin_x` is False.
    auto_label : bool
        If True, automatically fills in missing labels and title with defaults.
    axis_labels : Optional[List[str]]
        A list of axis labels. If provided, must contain three strings.
        - For dual Y-axis: [primary x-axis label, primary y-axis label, secondary y-axis label].
        - For dual X-axis: [primary x-axis label, primary y-axis label, secondary x-axis label].
    plot_title : Optional[str]
        Title of the plot. If None and `auto_label` is True, a default title will be used.
    use_twin_x : bool
        If True, indicates a dual Y-axis plot (X1 vs Y1 and X1 vs Y2).
        If False, indicates a dual X-axis plot (X1 vs Y1 and X2 vs Y1).

    Returns
    -------
    Tuple[str, str, str, str, List[str]]
        A tuple containing:
        - x1y1_label: The finalized label for the primary plot.
        - x1y2_label: The finalized label for the secondary Y-axis plot.
        - x2y1_label: The finalized label for the secondary X-axis plot.
        - plot_title: The finalized plot title.
        - axis_labels: The finalized axis labels as a list of three strings.

    Notes
    -----
    - If `auto_label` is True, missing labels are replaced with default values:
        - Default axis labels:
            - Dual Y-axis: ['X', 'Y1', 'Y2']
            - Dual X-axis: ['X1', 'Y', 'X2']
        - Default data labels:
            - Dual Y-axis: ['X1 vs Y1', 'X1 vs Y2']
            - Dual X-axis: ['Y vs X1', 'X2 vs Y1']
        - Default title: 'Plot'.
    - If `auto_label` is False and labels are missing, they are replaced with empty strings.

    Examples
    --------
    >>> dual_axes_label_management(None, None, None, True, None, None, True)
    ('X1 vs Y1', 'X1 vs Y2', '', 'Plot', ['X', 'Y1', 'Y2'])

    >>> dual_axes_label_management("Primary", None, None, False, ['Time', 'Value', 'Frequency'], "Custom Plot", False)
    ('Primary', '', '', 'Custom Plot', ['Time', 'Value', 'Frequency'])
    """
    # Set defaults for axis labels and data labels based on `use_twin_x`
    default_axis_labels = ['X', 'Y1', 'Y2'] if use_twin_x else ['X1', 'Y', 'X2']
    default_data_labels = ['X1 vs Y1', 'X1 vs Y2'] if use_twin_x else ['Y vs X1', 'X2 vs Y1']
    default_title = 'Plot'
    
    if auto_label:
        # Use defaults for missing labels
        axis_labels = axis_labels or default_axis_labels
        x1y1_label = x1y1_label or default_data_labels[0]
        if use_twin_x:
            x1y2_label = x1y2_label or default_data_labels[1]
        else:
            x2y1_label = x2y1_label or default_data_labels[1]
        plot_title = plot_title or default_title
    else:
        # If auto_label is False, use provided labels or leave blank
        axis_labels = axis_labels or ['', '', '']
        x1y1_label = x1y1_label or ''
        x1y2_label = x1y2_label or ''
        x2y1_label = x2y1_label or ''
        plot_title = plot_title or ''
    
    return x1y1_label, x1y2_label, x2y1_label, plot_title, axis_labels



"""Created on Jun 12 13:49:07 2024"""

import numpy as np


def transpose1d(array: np.ndarray) -> np.ndarray:
    """
    Transposes a given array.

    Parameters
    ----------
    array:
        Given numpy array.

    Returns
    -------
    np.ndarray
        Transposed numpy array
    """
    return np.array([array]).transpose() if len(array.shape) == 1 else array.transpose()


def reshape_with_padding(array: np.ndarray, new_shape: tuple, pad_value: int = 0) -> np.ndarray:
    """
    Reshape an array to a new shape with padding if necessary.

    Parameters
    ----------
    array : np.ndarray
        The input array to be reshaped.
    new_shape : tuple of int
        The desired shape for the output array.
    pad_value : scalar, optional
        The value to use for padding if the new shape requires more elements
        than the original array contains. The default is 0.

    Returns
    -------
    np.ndarray
        The reshaped array with padding applied if necessary.
    """
    new_array = np.full(new_shape, pad_value)
    new_array.flat[:len(array)] = array

    return new_array


def moving_average(array: np.ndarray, window_size: int) -> np.ndarray:
    """
    Compute the moving average of a given 1D array.

    Parameters
    ----------
    array : np.ndarray
        The 1-D array for which the moving average is to be computed.
    window_size : int
        The size of the moving window.

    Returns
    -------
    np.ndarray
        An array containing the moving averages. The length of this array will be `len(array) - window_size + 1`.
    """
    return np.convolve(array, np.ones(window_size), 'valid') / window_size



"""Created on Jul 23 17:17:17 2022."""


class EzFileErrs(Exception):
    """Base class for exceptions in EzFile."""
    pass


class LineNumberOutOfBounds(EzFileErrs):
    """Exception raised when a line number is out of bounds."""
    pass



"""Created on Jul 23 18:38:52 2022."""

from .eIO import LineNumberOutOfBounds


def check_for_errors(open_file: list, lines_to_read: list):
    """
    Checks if the requested lines to read are within the valid range for the file.

    Parameters
    ----------
    open_file : list
        A list of file content lines.
    lines_to_read : list
        A list of line numbers requested to be read.

    Raises
    ------
    LineNumberOutOfBounds
        If any line number is out of the bounds of the file.
    """

    file_length = len(open_file)
    err = False

    # Check if all line numbers are non-negative
    if all(line >= 0 for line in lines_to_read):
        if max(lines_to_read) >= file_length:
            err = True

    # Check if any negative line numbers are present
    elif any(line < 0 for line in lines_to_read):
        if abs(min(lines_to_read)) > file_length:
            err = True
        else:
            # Adjust negative indices to correspond to positive ones (from the end)
            for idx, line in enumerate(lines_to_read):
                if line < 0:
                    lines_to_read[idx] += file_length

    if err:
        raise LineNumberOutOfBounds("The line number specified is outside the line numbers of the file.")



"""Created on Jul 20 12:21:33 2022."""


class EzListErrs(Exception):
    pass


class AlphabetFound(EzListErrs):
    pass


class IndexOutOfList(EzListErrs):
    pass


class UnequalElements(EzListErrs):
    pass


class GotAnUnknownValue(EzListErrs):
    pass


class ChildListLengthError(EzListErrs):
    pass


class StringNotPassed(EzListErrs):
    pass


class InvalidInputParameter(EzListErrs):
    pass



"""Created on Aug 17 23:51:58 2022."""


def change_value_to_list(input_dictionary):
    for key, value in input_dictionary.items():
        if not isinstance(value, list):
            input_dictionary[key] = [input_dictionary[key]]

    return input_dictionary


def change_list_to_values(input_dictionary):
    for key, value in input_dictionary.items():
        if len(input_dictionary[key]) == 1:
            input_dictionary[key] = input_dictionary[key][0]

    return input_dictionary


class PrettyPrint:
    def __init__(self, input_dictionary: dict):
        self.inp_dict = input_dictionary

    def __get_max_width(self):
        value_widths = [len(str(value)) for value in self.inp_dict.values()]
        max_width = max(value_widths)
        return max(max_width + 1, 71) if max_width % 2 == 0 else max(max_width, 71)

    def __str__(self):
        max_width = self.__get_max_width()

        width = (max_width - 1) // 2 - 1
        pline = '-' * max_width + '\n'

        out = pline
        out += f"|{'dict_key'.center(width)}|{'dict_value'.center(width)}|\n"
        out += pline
        out += '\n'.join([f"|{k.center(width)}|{v.center(width)}|" for k, v in self.inp_dict.items()]) + '\n'
        out += pline

        return out



"""Created on Jul 20 00:17:54 2022."""

import itertools
import os
from typing import Union

from .eOS import FileNotPresent


class GetFiles:

    def __init__(self, input_variable, var_type, working_directory=os.curdir):
        self.input_variable = input_variable
        self.w_dir = working_directory
        self.var_type = var_type

        self.lof = None

        self.__initialize_lof()

    def __initialize_lof(self):
        """Get the list of files from the input folder with the given extension(s)."""
        if isinstance(self.input_variable, str):
            self.input_variable = (self.input_variable,)

        if self.var_type == 'ext':
            lof_ = ((f for f in os.listdir(self.w_dir) if f.endswith(x) and not os.path.isdir(f))
                    for x in self.input_variable)
        else:
            lof_ = ((f for f in os.listdir(self.w_dir) if x in f and not os.path.isdir(f)) for x in
                    self.input_variable)

        # taken from https://stackoverflow.com/a/953097/3212945
        self.lof = list(itertools.chain.from_iterable(lof_))

    def exclude(self, exclude_file: Union[str, list]):
        """
        Exclude specific files from the obtained list of files.

        Parameters
        ----------
        exclude_file : Union[str, list]
            File name or list of files to exclude from the obtained list of files.
        """
        if isinstance(exclude_file, str):
            exclude_file = (exclude_file,)

        mask_ = [x not in self.lof for x in exclude_file]

        if True in mask_:
            raise FileNotPresent(f'The file(s) named '
                                 f'{", ".join(itertools.compress(exclude_file, mask_))} '
                                 f'do not exist in the list of files.')
        else:
            tuple(self.lof.remove(x) for x in exclude_file)

    def sort(self, reverse: bool = False):
        """
        Sort the list of files obtained.

        Parameters
        ----------
        reverse : bool, optional
            Whether to reverse the sorting order of the list of files or not. The default is False.
        """
        self.lof.sort(reverse=reverse)

    @property
    def list_(self) -> list:
        """
        Show the list of files obtained.

        Returns
        -------
        list
            List of files matching the input extension.
        """
        return self.lof



"""Created on Jul 18 23:26:48 2022."""

import os
import shutil
from typing import Union

from .backend.uOS import GetFiles


class ListOfFilesFromExtensions(GetFiles):
    """Class for getting the list of files from a folder."""

    def __init__(self, extension: Union[str, list], directory: str = os.curdir):
        """
        Initialization method for ListOfFiles class.

        Parameters
        ----------
        extension : Union[str, list]
            The type of file to be picked from the working_directory.
        directory : str
            The directory from where the files are to be picked.

        Examples
        ----------
        Suppose we want to pick all `.py` files from the current folder,

        >>> list_of_files = ListOfFilesFromExtensions(extension='.py')

        To see the selected files,

        >>> list_of_files.list_

        To select multiple types of files, just pass in a list

        >>> list_of_files = ListOfFilesFromExtensions(extension=['.py', '.txt'])

        Now, let's say we want to remove a `remove_me.py` from the `list_of_files`

        >>> list_of_files.exclude(exclude_file='remove_me.py')
        """
        super(ListOfFilesFromExtensions, self).__init__(input_variable=extension,
                                                        var_type='ext',
                                                        working_directory=directory)


class ListOfFilesFromName(GetFiles):
    """Class for getting the list of files from a folder."""

    def __init__(self, file_name: Union[str, list], directory: str = os.curdir):
        """
        Initialization method for ListOfFiles class.

        Parameters
        ----------
        file_name : Union[str, list]
            The name of file to be picked from the working_directory. The file is picked with
            `in` keyword.
        directory : str
            The directory from where the files are to be picked.

        Examples
        ----------
        Suppose we want to pick all files from the current folder with contianing `main` in their
        names,

        >>> list_of_files = ListOfFilesFromName(file_name='main')

        To see the selected files,

        >>> list_of_files.list_

        To select multiple files with different names,

        >>> list_of_files = ListOfFilesFromName(file_name=['main', 'test'])

        Now, let's say we want to remove a `remove_me.py` from the `list_of_files`

        >>> list_of_files.exclude(exclude_file='remove_me.py')
        """
        super(ListOfFilesFromName, self).__init__(input_variable=file_name,
                                                  var_type='name',
                                                  working_directory=directory)


def move_directory_contents(old_path: str, new_path: str):
    """
    Moves the contents from `old_path` to `new_path`.

    Parameters
    ----------
    old_path:
        The old directory from which the files are to be moved.
    new_path:
        The new directory to which the files are to be moved.
    """
    if not os.path.exists(old_path):
        raise Exception(f"The source directory '{old_path}' does not exist.")

    if not os.path.exists(new_path):
        os.makedirs(new_path)

    for item in os.listdir(old_path):
        old_item_path = os.path.join(old_path, item)
        new_item_path = os.path.join(new_path, item)
        shutil.move(old_item_path, new_item_path)

    # Optionally, remove the old directory if it's empty
    if not os.listdir(old_path):
        os.rmdir(old_path)
    else:
        print(f"Warning: '{old_path}' is not empty after move.")



"""Created on Jul 23 23:41:18 2024"""

__all__ = ['plot_two_column_file', 'plot_xy', 'plot_xyy', 'plot_with_dual_axes', 'two_subplots', 'n_plotter']

from typing import List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import rcParams

from .backend import ePlotting as ePl, uPlotting as uPl

# safeguard
line_plot = "LinePlot"
scatter_plot = "ScatterPlot"
plot_dictionary_type = Optional[Union[uPl.LinePlot, uPl.ScatterPlot]]
axis_return = Union[List[plt.axis], plt.axis]


def plot_two_column_file(file_name: str,
                         delimiter: str = ',',
                         skip_header: bool = False,
                         data_label: str = 'X vs Y',
                         auto_label: bool = False,
                         is_scatter: bool = False,
                         plot_dictionary: Optional[plot_dictionary_type] = None,
                         axis: Optional[plt.axis] = None) -> axis_return:
    """
    Reads a two-column file (x, y) and plots the data.

    This function reads a file containing two columns (e.g., x and y values) and plots them
    using either a line plot or scatter plot based on the user's preference.

    Parameters
    ----------
    file_name : str
        The path to the file to be plotted. The file should contain two columns (x and y data).
    delimiter : str, optional
        The delimiter used in the file (default is ',').
    skip_header: bool, optional
        If True, skips the first row in the given data file, otherwise does nothing. Default is False.
    data_label: str
        Data label for the plot to put in the legend. Defaults to 'X vs Y'.
    auto_label : bool, optional
        If True, automatically sets the x-axis label, y-axis label, and plot title. Default is False.
    is_scatter : bool, optional
        If True, creates a scatter plot. Otherwise, creates a line plot. Default is False.
    plot_dictionary: Union[LinePlot, ScatterPlot], optional
        An object representing the plot data, either a `LinePlot` or `ScatterPlot`,  to be passed to the matplotlib plotting library.
         If None, a default plot type will be used.
    axis: Optional[plt.axis]
        The axis object to draw the plots on. If not passed, a new axis object will be created internally.

    Returns
    -------
    list:
        Either a single or double axis list.
    """
    # CHANGELIST:
    #   - Removed `fig_size` and added `data_label` parameter
    data = np.genfromtxt(file_name, delimiter=delimiter, skip_header=skip_header)
    
    if data.shape[1] != 2:
        raise ValueError("The file must contain exactly two columns of data.")
    
    x_data, y_data = data.T
    
    return plot_with_dual_axes(x1_data=x_data, y1_data=y_data, x1y1_label=data_label, auto_label=auto_label,
                               is_scatter=is_scatter, plot_dictionary=plot_dictionary, axis=axis)


def plot_xy(x_data: np.ndarray, y_data: np.ndarray,
            data_label: str = 'X vs Y',
            auto_label: bool = False, is_scatter: bool = False,
            plot_dictionary: Optional[plot_dictionary_type] = None,
            axis: Optional[plt.axis] = None) -> axis_return:
    """
    Plots x_data against y_data with customizable options.

    This function accepts two arrays of data (x and y) and plots them using either
    a line plot or scatter plot, with options for labels and figure size.

    Parameters
    ----------
    x_data : np.ndarray
        The data for the x-axis.
    y_data : np.ndarray
        The data for the y-axis.
    data_label: str
        Data label for the plot to put in the legend. Defaults to 'X vs Y'.
    auto_label : bool, optional
        If True, automatically sets x and y-axis labels and the plot title. Default is False.
    is_scatter : bool, optional
        If True, creates a scatter plot. Otherwise, creates a line plot. Default is False.
    plot_dictionary: Union[LinePlot, ScatterPlot], optional
        An object representing the plot data, either a `LinePlot` or `ScatterPlot`,  to be passed to the matplotlib plotting library.
         If None, a default plot type will be used.
    axis: Optional[plt.axis]
        The axis object to draw the plots on. If not passed, a new axis object will be created internally.

    Returns
    -------
    list:
        Either a single or double axis list.
    """
    # CHANGELIST:
    #   - Removed `fig_size` parameter
    return plot_with_dual_axes(x1_data=x_data, y1_data=y_data, x1y1_label=data_label, auto_label=auto_label,
                               is_scatter=is_scatter, plot_dictionary=plot_dictionary, axis=axis)


def plot_xyy(x_data: np.ndarray, y1_data: np.ndarray, y2_data: np.ndarray,
             data_labels: Optional[List[str]] = None, auto_label: bool = False,
             is_scatter: bool = False, plot_dictionary: plot_dictionary_type = None,
             axis: Optional[plt.Axes] = None) -> plt.Axes:
    """
    Plot two sets of y-data (`y1_data` and `y2_data`) against the same x-data (`x_data`) on the same plot.

    Parameters
    ----------
    x_data : np.ndarray
        The x-axis data for both plots.
    y1_data : np.ndarray
        The first set of y-axis data to be plotted against `x_data`.
    y2_data : np.ndarray
        The second set of y-axis data to be plotted against `x_data`.
    data_labels : list of str, optional
        The labels for the two datasets. Default is `['X vs Y1', 'X vs Y2']`.
    auto_label : bool, optional
        Whether to automatically label the plot. Default is `False`.
    is_scatter : bool, optional
        Whether to create a scatter plot (`True`) or a line plot (`False`). Default is `False`.
    plot_dictionary : dict, optional
        A dictionary containing plot configuration parameters for the two datasets. Default is `None`.
    axis : plt.Axes, optional
        A Matplotlib axis to plot on. If `None`, a new axis is created. Default is `None`.

    Returns
    -------
    plt.Axes
        The axis object containing the plotted data.
    """
    # CHANGELIST:
    #   - Removed `fig_size` parameter
    data_labels = ['X vs Y1', 'X vs Y2'] if data_labels is None else data_labels
    plot_config_1, plot_config_2 = uPl.split_dictionary(plot_dictionary)
    
    axis = plot_with_dual_axes(x1_data=x_data, y1_data=y1_data, x1y1_label=data_labels[0], auto_label=auto_label,
                               is_scatter=is_scatter, plot_dictionary=plot_config_1, axis=axis)
    return plot_with_dual_axes(x1_data=x_data, y1_data=y2_data, x1y1_label=data_labels[1], auto_label=auto_label,
                               is_scatter=is_scatter, plot_dictionary=plot_config_2, axis=axis)


def plot_with_dual_axes(x1_data: np.ndarray, y1_data: np.ndarray,
                        x2_data: np.ndarray = None, y2_data: np.ndarray = None,
                        x1y1_label: str = None,
                        x1y2_label: str = None,
                        x2y1_label: str = None,
                        use_twin_x: bool = False,
                        auto_label: bool = False,
                        axis_labels: List[str] = None,
                        plot_title: str = None,
                        is_scatter: bool = False,
                        plot_dictionary: Optional[plot_dictionary_type] = None,
                        axis: Optional[plt.axis] = None) -> axis_return:
    """
    Plots data with options for dual axes (x or y) or single axis.

    Parameters
    ----------
    x1_data : np.ndarray
        Data for the primary x-axis.
    y1_data : np.ndarray
        Data for the primary y-axis.
    x2_data : np.ndarray, optional
        Data for the secondary x-axis (used for dual x-axis plots).
    y2_data : np.ndarray, optional
        Data for the secondary y-axis (used for dual y-axis plots).
    x1y1_label : str, optional
        Label for the plot of X1 vs Y1. If None and `auto_label` is True, defaults to 'X1 vs Y1'.
    x1y2_label : str, optional
        Label for the plot of X1 vs Y2 (when using dual Y-axes). If None and `auto_label` is True, defaults to 'X1 vs Y2'.
    x2y1_label : str, optional
        Label for the plot of X2 vs Y1 (when using dual X-axes). If None and `auto_label` is True, defaults to 'X2 vs Y1'.
    use_twin_x : bool, optional
        If True, creates dual y-axis plot. If False, creates dual x-axis plot. Default is False.
    auto_label : bool, optional
        If True, automatically assigns labels if none are provided. Default is False.
    axis_labels : list of str, optional
        List of axis labels in the form [xlabel, ylabel1, ylabel2]. If None and `auto_label` is True, defaults to ['X', 'Y1', 'Y2'] or ['X1', 'Y', 'X2'].
    plot_title : str, optional
        Title of the plot. If None and `auto_label` is True, defaults to 'Plot'. If None and `auto_label` is False, no title is added.
    is_scatter : bool, optional
        If True, creates scatter plot; otherwise, line plot. Default is False.
    plot_dictionary: Union[LinePlot, ScatterPlot], optional
        An object representing the plot data, either a `LinePlot` or `ScatterPlot`, to be passed to the matplotlib plotting library.
    axis: Optional[plt.axis]
        The axis object to draw the plots on. If not passed, a new axis object will be created internally.

    Returns
    -------
    list:
        Either a single or double axis list.
    """
    # CHANGELIST:
    #   - Works with axis
    #   - Works with default `fig_size` if no figure size is specified
    #   - Simplified the workings of data labels
    #   - Removed `fig_size` parameter
    #   - Handles `auto_label`, `axis_labels` and `plot_title` separately
    #   - Handles empty labels correctly as well
    #   - Can deal with labels and data validations
    
    labels = uPl.dual_axes_label_management(x1y1_label=x1y1_label, x1y2_label=x1y2_label, x2y1_label=x2y1_label,
                                            auto_label=auto_label, axis_labels=axis_labels, plot_title=plot_title,
                                            use_twin_x=use_twin_x)
    
    x1y1_label, x1y2_label, x2y1_label, plot_title, axis_labels = labels
    
    uPl.dual_axes_data_validation(x1_data=x1_data, x2_data=x2_data, y1_data=y1_data, y2_data=y2_data,
                                  use_twin_x=use_twin_x, axis_labels=axis_labels)
    
    if axis:
        ax1 = axis
    else:
        _, ax1 = plt.subplots(figsize=rcParams["figure.figsize"])
    
    plot_items = uPl.plot_dictionary_handler(plot_dictionary=plot_dictionary)
    dict1 = {key: (value[0] if isinstance(value, list) else value) for key, value in plot_items}
    uPl.plot_or_scatter(axes=ax1, scatter=is_scatter)(x1_data, y1_data, label=x1y1_label, **dict1)
    
    ax2 = None
    ax1.set_xlabel(axis_labels[0])
    ax1.set_ylabel(axis_labels[1])
    if plot_title:
        ax1.set_title(plot_title)
    
    if use_twin_x:
        ax2 = ax1.twinx()
        if y2_data is not None:
            dict2 = {key: (value[1] if len(value) > 1 else None) for key, value in plot_items}
            uPl.plot_or_scatter(axes=ax2, scatter=is_scatter)(x1_data, y2_data, label=x1y2_label, **dict2)
            ax2.set_ylabel(axis_labels[2])

    elif x2_data is not None:
        ax2 = ax1.twiny()
        dict2 = {key: (value[1] if len(value) > 1 else None) for key, value in plot_items}
        uPl.plot_or_scatter(axes=ax2, scatter=is_scatter)(x2_data, y1_data, label=x2y1_label, **dict2)
        ax2.set_xlabel(axis_labels[2])
    
    if x1y1_label or x1y2_label or x2y1_label:
        handles, labels = ax1.get_legend_handles_labels()
        if ax2:
            handles2, labels2 = ax2.get_legend_handles_labels()
            handles += handles2
            labels += labels2
        ax1.legend(handles, labels, loc='best')
    
    plt.tight_layout()
    
    return (ax1, ax2) if ax2 else ax1


def two_subplots(x_data: List[np.ndarray], y_data: List[np.ndarray],
                 x_labels: List[str], y_labels: List[str], data_labels: List[str],
                 orientation: str = 'h',
                 auto_label: bool = False,
                 is_scatter: bool = False,
                 subplot_dictionary: Optional[uPl.SubPlots] = None,
                 plot_dictionary: Optional[Union[uPl.LinePlot, uPl.ScatterPlot]] = None) -> None:
    """
    Creates two subplots arranged horizontally or vertically, with optional customization.

    This function internally calls `n_plotter` to handle the plotting of each subplot.
    `n_plotter` arranges the subplots and applies relevant plot and subplot dictionaries.

    Parameters
    ----------
    x_data : list of np.ndarray
        List containing x-axis data arrays for each subplot.
    y_data : list of np.ndarray
        List containing y-axis data arrays for each subplot.
    x_labels : list of str
        List of labels for the x-axes in each subplot.
    y_labels : list of str
        List of labels for the y-axes in each subplot.
    data_labels : list of str
        List of labels for the data series in each subplot.
    orientation : str, optional, default='h'
        Orientation of the subplots, either 'h' for horizontal or 'v' for vertical.
    auto_label : bool, default False
        Automatically assigns labels to subplots if `True`.
    is_scatter : bool, default False
        If `True`, plots data as scatter plots; otherwise, plots as line plots.
    subplot_dictionary : dict, optional
        Dictionary of parameters for subplot configuration.
    plot_dictionary : LinePlot or ScatterPlot, optional
        Object containing plot styling parameters. Defaults to `LinePlot`.
    """
    # CHANGELIST:
    #   - Supports passing two x and two y arguments for enhanced flexibility.
    #   - Added support for subplot dictionaries; testing needed for LinePlot/ScatterPlot cases.
    #   - Allows X and Y data to be passed as lists for easier data handling.
    #   - For `two_subplots`, enables horizontal or vertical orientation (since there are only two subplots).
    #   - Falls back to default behavior if no subplot dictionary is provided.
    #   - Defaults to standard plot settings if no plot dictionary is provided.
    #   - Includes an `is_scatter` option for toggleable scatter plot functionality.
    #   - Now supports plot dictionaries with multiple parameters; if the second parameter is missing, the first is applied to both subplots.
    #   - Returns the axes object for better integration with other plotting functions.
    #   - Adapts to `n_plotter` for enhanced plot flexibility.
    #   - Removed the redundant `axes` variable for a cleaner implementation.
    
    if orientation == 'h':
        n_rows, n_cols = 1, 2
    elif orientation == 'v':
        n_rows, n_cols = 2, 1
    else:
        raise ePl.OrientationError("The orientation must be either \'h\' or \'v\'.")
    
    return n_plotter(x_data=x_data, y_data=y_data,
                     n_rows=n_rows, n_cols=n_cols,
                     x_labels=x_labels, y_labels=y_labels, data_labels=data_labels,
                     auto_label=auto_label, is_scatter=is_scatter,
                     subplot_dictionary=subplot_dictionary, plot_dictionary=plot_dictionary)


def n_plotter(x_data: List[np.ndarray], y_data: List[np.ndarray],
              n_rows: int, n_cols: int,
              x_labels: Optional[List[str]] = None, y_labels: Optional[List[str]] = None,
              data_labels: Optional[List[str]] = None,
              auto_label: bool = False,
              is_scatter: bool = False,
              subplot_dictionary: Optional[uPl.SubPlots] = None,
              plot_dictionary: Optional[Union[uPl.LinePlot, uPl.ScatterPlot]] = None) -> Union[plt.figure, plt.axis]:
    """
    Plots multiple subplots in a grid with optional customization for each subplot.

    Parameters
    ----------
    x_data : list of np.ndarray
        List of x-axis data arrays for each subplot.
    y_data : list of np.ndarray
        List of y-axis data arrays for each subplot.
    n_rows : int
        Number of rows in the subplot grid.
    n_cols : int
        Number of columns in the subplot grid.
    x_labels : list of str, optional
        List of labels for the x-axes of each subplot.
    y_labels : list of str, optional
        List of labels for the y-axes of each subplot.
    data_labels : list of str, optional
        List of labels for the data series in each subplot.
    auto_label : bool, default False
        Automatically assigns labels to subplots if `True`.
    is_scatter : bool, default False
        If `True`, plots data as scatter plots; otherwise, plots as line plots.
    subplot_dictionary : dict, optional
        Dictionary of parameters for subplot configuration.
    plot_dictionary : LinePlot or ScatterPlot, optional
        Object containing plot styling parameters. Defaults to `LinePlot`.
    """
    # CHANGELIST:
    #   - Simplified handling of data for subplots with n_rows x n_cols layout (n_cols > n_rows).
    #   - Replaced `plot_on_dual_axes` with `plot_xy` for better flexibility in handling data.
    #   - Simplified logic for labels and legends, improving clarity and usability.
    #   - Improved dictionary handling for single-row subplots.
    #   - Removed the use of axes variable; now directly handles axes passed as arguments.
    #   - Optimized behavior for multi-row and multi-column subplots, with better layout management.
    #   - Improved subplot decoration for multi-row and multi-column cases (x/y labels, ticks, etc.).
    #   - If `share_y = True`, other y-axes are no longer shown to avoid clutter.
    #   - Removed `plot_on_dual_axes` or `plot_xy` dependency, instead uses simple plot/scatter functionality.
    
    sp_dict = subplot_dictionary.get() if subplot_dictionary else uPl.SubPlots().get()
    
    fig, axs = plt.subplots(n_rows, n_cols, **sp_dict)
    axs = axs.flatten()
    
    plot_items = uPl.plot_dictionary_handler(plot_dictionary=plot_dictionary)
    
    main_dict = [{key: value[c] for key, value in plot_items} for c in range(n_cols * n_rows)]
    
    x_labels, y_labels = uPl.label_handler(x_labels=x_labels, y_labels=y_labels, n_rows=n_rows, n_cols=n_cols,
                                           auto_label=auto_label)
    
    shared_y = sp_dict.get('sharey')
    shared_x1 = sp_dict.get('sharex')
    shared_x2 = len(axs) - int(len(axs) / n_rows if n_rows > n_cols else n_cols)
    for index, ax, j, k in zip(range(n_cols * n_rows), axs, x_labels, y_labels):
        label = f'{x_labels[index]} vs {y_labels[index]}' if data_labels is None else data_labels[index]
        uPl.plot_or_scatter(axes=ax, scatter=is_scatter)(x_data[index], y_data[index], label=label, **main_dict[index])
        if shared_x1:
            if not index < shared_x2:
                ax.set_xlabel(j)
        else:
            ax.set_xlabel(j)
        if not (shared_y and index % n_cols != 0):
            ax.set_ylabel(k)
        ax.legend(loc='best')
    
    plt.tight_layout()
    
    return fig, axs



"""Created on Jul 23 04:26:54 2022."""

from typing import Any, Dict

from .backend.uDict import PrettyPrint


def sort_dictionary(input_dictionary, sort_by='values', sort_reverse=False) -> dict:
    """
    Sorts dictionary either by value or by keys.

    Parameters
    ----------
    input_dictionary:
        The given dictionary to be sorted.
    sort_by:
        The parameter to sort the dictionary on, either 'values' or 'keys'. Defaults to 'values'.
    sort_reverse:
        Whether to reverse sort the dictionary or not.

    Returns
    -------
    dict:
        Sorted dictionary
    """
    def _get_key(x):
        return x[1] if sort_by == 'values' else x[0]

    return dict(sorted(input_dictionary.items(), key=_get_key, reverse=sort_reverse))


def merge_dictionaries(*dicts: Dict[str, Any]) -> Dict[str, Any]:
    """
    Merge multiple dictionaries into a single dictionary.

    Parameters
    ----------
    dicts : Dict[str, Any]
        The dictionaries to be merged.

    Returns
    -------
    Dict[str, Any]
        The merged dictionary.
    """
    merged_dict = {}

    for dict_ in dicts:
        for key, value in dict_.items():
            if key in merged_dict:
                if isinstance(merged_dict[key], list) and isinstance(value, list):
                    merged_dict[key].extend(value)
                else:
                    merged_dict[key] = [merged_dict[key], value]
            else:
                merged_dict[key] = value

    return merged_dict


def pretty_print(input_dictionary) -> PrettyPrint:
    """
    Pretty prints the given dictionary.

    Parameters
    ----------
    input_dictionary:
        The dictionary to be pretty-printed.

    Returns
    -------
    PrettyPrint:
        PrettyPrinted dictionary.
    """
    return PrettyPrint(input_dictionary)


def get_key_index(input_dictionary: dict, key: object) -> int:
    """
    Find the index of a key in a dictionary.

    Parameters
    ----------
    input_dictionary : dict
        The dictionary to search.
    key : object
        The key to find.

    Returns
    -------
    int
        The index of the key in the dictionary.

    Examples
    --------
    >>> inp_dictionary = {'a': 1, 'b': 2, 'c': 3}
    >>> key_to_find = 'b'
    >>> get_key_index(inp_dictionary, key_to_find)
    1
    """
    keys = list(input_dictionary.keys())
    return keys.index(key)



"""Created on Jun 12 13:49:07 2022"""

import unittest

from ..mpyez import ezList
from ..mpyez.backend import uList
from ..mpyez.backend.eList import AlphabetFound, UnequalElements


class Test(unittest.TestCase):
    def test_numeric_list_to_string(self):
        self.assertEqual(uList.numeric_list_to_string([1, 2, 3]), ['1', '2', '3'])

    def test_string_list_to_numeric(self):
        self.assertEqual(ezList.string_list_to_numeric(['1', '2', '3']), [1, 2, 3])

        with self.assertRaises(AlphabetFound):
            ezList.string_list_to_numeric(['A'])

    def test_nested_list_to_list(self):
        self.assertEqual(ezList.nested_list_to_list([[1, 2, 3, 4], [5, 6, 7, 8]]),
                         [1, 2, 3, 4, 5, 6, 7, 8])

    def test_list_to_nested_list(self):
        self.assertEqual(ezList.list_to_nested_list([1, 2, 3, 4, 5, 6], 2),
                         [[1, 2], [3, 4], [5, 6]])
        self.assertEqual(ezList.list_to_nested_list([1, 2, 3, 4, 5, 6], 3),
                         [[1, 2, 3], [4, 5, 6]])

    def test_join_lists(self):
        self.assertEqual(ezList.join_lists([[1, 2, 3], [5, 6, 4]]),
                         [1, 2, 3, 5, 6, 4])

        self.assertEqual(ezList.join_lists([[1, 2, 3], [5, 6, 4]], sort=True),
                         [1, 2, 3, 4, 5, 6])

    def test_is_contained(self):
        a, b = [1, 2, 3], [1, 2, 3, 4]

        self.assertEqual(ezList.is_contained(a, b), True)
        self.assertEqual(ezList.is_contained(b, a), False)

    def test_Replace__single_index(self):
        inp_, ind_, wth_ = [1, 2, 3, 4, 5], 0, 10

        self.assertEqual(ezList.replace_at_index(inp_, ind_, wth_), [10, 2, 3, 4, 5])

    def test_Replace__multi_index(self):
        inp_, ind_, wth_ = [1, 2, 3, 4, 5], [0, 2, 3], [10, 10, 10]

        self.assertEqual(ezList.replace_at_index(inp_, ind_, wth_), [10, 2, 10, 10, 5])

    def test_Replace__single_index__multiple_values(self):
        inp_, ind_, wth_ = [1, 2, 3, 4, 5], 0, [10, 10, 10]

        with self.assertRaises(UnequalElements):
            ezList.replace_at_index(inp_, ind_, wth_)

    def test_Replace__multi_value(self):
        inp_, val_, wth_ = [1, 2, 3, 4, 5], [2, 3], [10, 12]

        self.assertEqual(ezList.replace_with_value(inp_, val_, wth_), [1, 10, 12, 4, 5])



"""Created on Jul 20 11:54:27 2022."""

import copy
from collections import Counter
from itertools import chain, compress
from typing import Any, List, Union

from .backend import eList as eL
from .backend.uList import CountObjectsInList, Replace


def equal_lists(lists: list) -> bool:
    """
    Check if all lists within a list have the same length.

    Parameters
    ----------
    lists : list of lists
        A list containing sublists to be checked.

    Returns
    -------
    bool
        True if all sublists have the same length, False otherwise.

    Examples
    --------
    >>> equal_lists([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    True
    >>> equal_lists([[1, 2, 3], [4, 5], [7, 8, 9]])
    False
    """
    return len(set(map(len, lists))) == 1


def string_list_to_numeric(str_list: list) -> list:
    """
    Convert all elements of a string lists to numeric.

    Parameters
    ----------
    str_list : list
        A list containing string elements.

    Raises
    ------
    errors.AlphabetFound
        This error is raised when the string is alphabetical in nature and not numerical.

    Returns
    -------
    list
        A list containing numeric elements.
    """
    mask_ = [element.isalpha() for element in str_list]

    if any(mask_):
        raise eL.AlphabetFound(f'An alphabet found in the list passed, '
                               f'{", ".join(compress(str_list, mask_))} cannot be processed.')

    return list(map(int, str_list))


def nested_list_to_list(nested_list: List[Any]) -> list:
    """
    Convert nested list to a single list.

    Parameters
    ----------
    nested_list : List[Any]
        An even/uneven nested list.

    Returns
    -------
    list
        A 1D list.

    Examples
    --------
    To convert

    >>> a = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]

    to a simple 1D list,

    >>> nested_list_to_list(a)
    >>> [1, 2, 3, 4, 5, 6, 7, 8, 9]

    Uneven lists can also be converted to 1D lists,

    >>> a = [[1, 2, 3], [4, 5], [6]]
    >>> nested_list_to_list(a)
    >>> [1, 2, 3, 4, 5, 6]
    """
    return list(chain.from_iterable(nested_list))


def list_to_nested_list(input_list: list, n_elements: int) -> List[list]:
    """
    Convert a single list to nested list.

    Parameters
    ----------
    input_list : list
        A simple, single list.
    n_elements : int
        Number of elements in each list.

    Returns
    -------
    List[list]
        A nested list with the n_elements per inner list.

    Notes
    -----
    For lists with, `len(list) % n_elements > 0`, the number of elements in the last inner list
    will not match the rest.

    Examples
    --------
    To convert,

    >>> a = [1, 2, 3, 4, 5, 6, 7, 8, 9]

    into a nested list with 3 elements per list,

    >>> list_to_nested_list(a, 3)
    >>> [[1, 2, 3], [4, 5, 6], [7, 8, 9]]

    Similarly, for the nested list to have four elements per inner list,

    >>> list_to_nested_list(a, 4)
    >>> [[1, 2, 3, 4], [5, 6, 7, 8], [9]]
    """
    return [input_list[i:i + n_elements] for i in range(0, len(input_list), n_elements)]


def join_lists(input_lists: List[Any], get_unique: bool = False, sort: bool = False) -> List[Any]:
    """
    Joins two or more lists.

    Parameters
    ----------
    input_lists : List[Any]
        A nested list with lists, tuples, ints, or floats in it.
    get_unique: bool
        Whether the output should contain unique values or not. The default if False.
    sort : bool
        Whether the output should be sorted or not. The default is False.

    Returns
    -------
    List[Any]
        A merger of all the input lists.
    """
    # taken from https://www.geeksforgeeks.org/extending-list-python-5-different-ways/
    out_list = list(chain([], *input_lists))

    # get indices of all list elements with type == list
    mask1 = [i for i, v in enumerate(out_list) if isinstance(v, list)]

    # change the values at mask1 indices from lists to tuples
    if mask1 is not None:
        for value in mask1:
            out_list[value] = tuple(out_list[value])

    # taken from https://stackoverflow.com/a/58666031/3212945
    if get_unique:
        unique = set()
        out_list = [value for value in out_list if not (value in unique or unique.add(value))]

    if sort:
        try:
            out_list = sorted(out_list)
        except TypeError:
            n_types = list(set([type(element) for element in out_list]))
            mask2 = [[index for index, value in enumerate(out_list) if isinstance(value, types)] for
                     types in n_types]

            out_list = [sorted(element) for element in
                        [[out_list[value] for value in index] for index in mask2]]
            out_list = nested_list_to_list(out_list)

    return out_list


def is_contained(child_list: list, parent_list: list) -> bool:
    """
    Check if the child_list is contained within the parent_list.

    Parameters
    ----------
    child_list : list
        The list to be checked for containment in parent list.
    parent_list : list
        The list to be checked for containment of child list.

    Returns
    -------
    bool
        Whether the child list is contained within parent list or not

    Examples
    --------
    To check whether,

    >>> a = [1, '2', 3, 'A']

    is contained within

    >>> b = [1, '2', 3, 4, 5, 'A']

    the is_contained method can be used

    >>> is_contained(a, b)
    >>> True
    """
    return all(child in parent_list for child in child_list)


def get_object_count(input_list: list, top_n: float = -1, get_tabular_form: bool = False):
    """
    Get the count of objects in a given list.

    Parameters
    ----------
    input_list : list
        The given list to get the element count from.
    top_n : float, optional
        Whether to show top `N` elements or all of them from the list. The default is -1.
    get_tabular_form : bool, optional
        DESCRIPTION. The default is False.

    Returns
    -------
    count_obj
        Either a dictionary or a CountObjectsInList object representing the counts of objects in a
        given list.

    """
    counts = dict(Counter(input_list))

    if get_tabular_form:
        obj_ = CountObjectsInList(counts)
        count_obj = obj_[:] if top_n == -1 else obj_ if top_n == 0 else obj_[0: top_n]
    else:
        count_obj = counts

    return count_obj


def sort_(input_list: list, ascending_order: bool = True, get_sorting_indices: bool = True) -> list:
    """
    Sort the given list.

    Parameters
    ----------
    input_list : list
        The list to be sorted.
    ascending_order : bool, optional
        Whether to sort the list in ascending or descending order. The default is False.
    get_sorting_indices : bool, optional
        Get the order of sort in the original list. The default is True.

    Returns
    -------
        Sorted list. Sort list indices, optional.

    """
    sort_order = False if ascending_order else True

    if get_sorting_indices:
        sorted_list = (list(i) for i in zip(*sorted(zip(input_list, range(len(input_list))))))

        return [element[::-1] for element in sorted_list] if sort_order else list(sorted_list)
    else:
        return sorted(input_list, reverse=sort_order)


def remove_(input_list: list, value_to_remove: Union[list, tuple, str, int],
            get_new_list: bool = False) -> list:
    """
    Remove a certain value from the input list.    

    Parameters
    ----------
    input_list : list
        The list from which the value is to be removed.
    value_to_remove : Union[list, tuple, str, int]
        The value to remove. The value can either be an int, str, tuple or even a nested list.
    get_new_list : bool, optional
        Whether the original list should be preserved or not. The default is False.

    Returns
    -------
    modified_list : list
        Modified list with the value removed from it.

    """
    modified_list = input_list if not get_new_list else copy.deepcopy(input_list)

    if isinstance(value_to_remove, (tuple, str, list)):
        ind_ = modified_list.index(value_to_remove)
        del modified_list[ind_]
    else:
        modified_list.pop(value_to_remove)

    return modified_list


def move_element_in_list(input_list: list, old_position: Union[list, int], new_position: Union[list, int],
                         get_new_list: bool = False) -> list:
    """
    Moves an element from `old_position` in the given list to `new_position`.

    Parameters
    ----------
    input_list : list
        The list in which the element/elements are to be moved.
    old_position : Union[list, int]
        The index (or list of index) on which the element to be moved is currently present.
    new_position : Union[list, int]
        The index (or list of index) to which the element is to be moved.
    get_new_list : bool, optional
        Whether the original list should be preserved or not. The default is False.

    Returns
    -------
    list :
        A list with the position of elements changed.

    """
    temp_ = input_list if not get_new_list else copy.deepcopy(input_list)

    temp_.insert(new_position, temp_.pop(old_position))

    return temp_


def replace_at_index(input_list: list, work_on: Union[list, int], replace_with: Union[list, int], new_list: bool = False):
    """
    Replace elements in `input_list` at specific indices.

    Parameters
    ----------
    input_list : list
        The original list to modify.
    work_on : Union[list, int]
        A single index or a list of indices where elements will be replaced.
    replace_with : Union[list, int]
        A single value or list of values that will replace the existing elements at the specified indices.
    new_list : bool, optional
        If True, returns a new list with the replacements. Otherwise, modifies the original list in place. Default is False.

    Returns
    -------
    list
        The modified list with replaced values.
    """
    return Replace(input_list, work_on, replace_with, new_list).at_index()


def replace_with_value(input_list: list, work_on: Union[list, int], replace_with: Union[list, int], new_list: bool = False):
    """
    Replace specific values in `input_list` with new values.

    Parameters
    ----------
    input_list : list
        The original list to modify.
    work_on : Union[list, int]
        A single value or list of values that will be replaced in the list.
    replace_with : Union[list, int]
        A single value or list of values that will replace the specified values.
    new_list : bool, optional
        If True, returns a new list with the replacements. Otherwise, modifies the original list in place. Default is False.

    Returns
    -------
    list
        The modified list with replaced values.
    """
    return Replace(input_list, work_on, replace_with, new_list, 'value').at_value()


def difference_between_lists(input_list1: list, input_list2: list):
    """
    Find the differences between two lists.

    Parameters
    ----------
    input_list1 : list
        The first list to compare.
    input_list2 : list
        The second list to compare.

    Returns
    -------
    tuple of lists
        Two lists representing the differences:
        - First list contains elements in `input_list1` not in `input_list2`.
        - Second list contains elements in `input_list2` not in `input_list1`.
    """
    diff1 = [element for element in input_list1 if element not in input_list2]
    diff2 = [element for element in input_list2 if element not in input_list1]
    return diff1, diff2


def index_(input_list: list, iterator: Union[list, int]):
    """
    Get the index or indices of specific values in `input_list`.

    Parameters
    ----------
    input_list : list
        The list to search for values.
    iterator : Union[list, int]
        A single value or list of values for which indices will be returned.

    Returns
    -------
    Union[int, list of int]
        The index of the value if `iterator` is a single value, or a list of indices if `iterator` is a list.
    """
    if isinstance(iterator, int):
        return input_list.index(iterator)
    else:
        return [input_list.index(elem) for elem in iterator]
